@@ -17,6 +17,7 @@
 #include "internal.h"
 #include "internal/enc.h"
 #include "internal/encoding.h"
+#include "internal/error.h"
 #include "internal/inits.h"
 #include "internal/load.h"
 #include "internal/object.h"
@@ -49,16 +50,14 @@ void rb_encdb_declare(const char *name);
 int rb_encdb_replicate(const char *name, const char *orig);
 int rb_encdb_dummy(const char *name);
 int rb_encdb_alias(const char *alias, const char *orig);
-void rb_encdb_set_unicode(int index);
 #pragma GCC visibility pop
 #endif
 
 static ID id_encoding;
 VALUE rb_cEncoding;
 
-#define DEFAULT_ENCODING_LIST_CAPA 128
-static VALUE rb_default_encoding_list;
-static VALUE rb_additional_encoding_list;
+#define ENCODING_LIST_CAPA 256
+static VALUE rb_encoding_list;
 
 struct rb_encoding_entry {
     const char *name;
@@ -67,12 +66,29 @@ struct rb_encoding_entry {
 };
 
 static struct enc_table {
-    struct rb_encoding_entry *list;
+    struct rb_encoding_entry list[ENCODING_LIST_CAPA];
     int count;
-    int size;
     st_table *names;
 } global_enc_table;
 
+static int
+enc_names_free_i(st_data_t name, st_data_t idx, st_data_t args)
+{
+    ruby_xfree((void *)name);
+    return ST_DELETE;
+}
+
+void
+rb_free_global_enc_table(void)
+{
+    for (size_t i = 0; i < ENCODING_LIST_CAPA; i++) {
+        xfree((void *)global_enc_table.list[i].enc);
+    }
+
+    st_foreach(global_enc_table.names, enc_names_free_i, (st_data_t)0);
+    st_free_table(global_enc_table.names);
+}
+
 static rb_encoding *global_enc_ascii,
                    *global_enc_utf_8,
                    *global_enc_us_ascii;
@@ -104,7 +120,7 @@ static rb_encoding *global_enc_ascii,
 static const rb_data_type_t encoding_data_type = {
     "encoding",
     {0, 0, 0,},
-    0, 0, RUBY_TYPED_FREE_IMMEDIATELY
+    0, 0, RUBY_TYPED_FREE_IMMEDIATELY | RUBY_TYPED_WB_PROTECTED
 };
 
 #define is_data_encoding(obj) (RTYPEDDATA_P(obj) && RTYPEDDATA_TYPE(obj) == &encoding_data_type)
@@ -128,47 +144,25 @@ enc_new(rb_encoding *encoding)
 static void
 enc_list_update(int index, rb_raw_encoding *encoding)
 {
-    if (index < DEFAULT_ENCODING_LIST_CAPA) {
-        VALUE list = rb_default_encoding_list;
-        if (list && NIL_P(rb_ary_entry(list, index))) {
-            /* initialize encoding data */
-            rb_ary_store(list, index, enc_new(encoding));
-        }
-    }
-    else {
-        RB_VM_LOCK_ENTER();
-        {
-            VALUE list = rb_additional_encoding_list;
-            if (list && NIL_P(rb_ary_entry(list, index))) {
-                /* initialize encoding data */
-                rb_ary_store(list, index - DEFAULT_ENCODING_LIST_CAPA, enc_new(encoding));
-            }
-        }
-        RB_VM_LOCK_LEAVE();
+    RUBY_ASSERT(index < ENCODING_LIST_CAPA);
+
+    VALUE list = rb_encoding_list;
+    if (list && NIL_P(rb_ary_entry(list, index))) {
+        /* initialize encoding data */
+        rb_ary_store(list, index, enc_new(encoding));
     }
 }
 
 static VALUE
 enc_list_lookup(int idx)
 {
-    VALUE list, enc;
+    VALUE list, enc = Qnil;
 
-    if (idx < DEFAULT_ENCODING_LIST_CAPA) {
-        if (!(list = rb_default_encoding_list)) {
-            rb_bug("rb_enc_from_encoding_index(%d): no rb_default_encoding_list", idx);
-        }
+    if (idx < ENCODING_LIST_CAPA) {
+        list = rb_encoding_list;
+        RUBY_ASSERT(list);
         enc = rb_ary_entry(list, idx);
     }
-    else {
-        RB_VM_LOCK_ENTER();
-        {
-            if (!(list = rb_additional_encoding_list)) {
-                rb_bug("rb_enc_from_encoding_index(%d): no rb_additional_encoding_list", idx);
-            }
-            enc = rb_ary_entry(list, idx - DEFAULT_ENCODING_LIST_CAPA);
-        }
-        RB_VM_LOCK_LEAVE();
-    }
 
     if (NIL_P(enc)) {
         rb_bug("rb_enc_from_encoding_index(%d): not created yet", idx);
@@ -210,7 +204,7 @@ check_encoding(rb_encoding *enc)
 {
     int index = rb_enc_to_index(enc);
     if (rb_enc_from_index(index) != enc)
-	return -1;
+        return -1;
     if (rb_enc_autoload_p(enc)) {
         index = rb_enc_autoload(enc);
     }
@@ -221,7 +215,7 @@ static int
 enc_check_encoding(VALUE obj)
 {
     if (!is_obj_encoding(obj)) {
-	return -1;
+        return -1;
     }
     return check_encoding(RDATA(obj)->data);
 }
@@ -231,7 +225,7 @@ static void
 not_encoding(VALUE enc)
 {
     rb_raise(rb_eTypeError, "wrong argument type %"PRIsVALUE" (expected Encoding)",
-	     rb_obj_class(enc));
+             rb_obj_class(enc));
 }
 
 static rb_encoding *
@@ -239,7 +233,7 @@ must_encoding(VALUE enc)
 {
     int index = enc_check_encoding(enc);
     if (index < 0) {
-	not_encoding(enc);
+        not_encoding(enc);
     }
     return DATA_PTR(enc);
 }
@@ -249,16 +243,16 @@ must_encindex(int index)
 {
     rb_encoding *enc = rb_enc_from_index(index);
     if (!enc) {
-	rb_raise(rb_eEncodingError, "encoding index out of bound: %d",
-		 index);
+        rb_raise(rb_eEncodingError, "encoding index out of bound: %d",
+                 index);
     }
     if (ENC_TO_ENCINDEX(enc) != (int)(index & ENC_INDEX_MASK)) {
-	rb_raise(rb_eEncodingError, "wrong encoding index %d for %s (expected %d)",
-		 index, rb_enc_name(enc), ENC_TO_ENCINDEX(enc));
+        rb_raise(rb_eEncodingError, "wrong encoding index %d for %s (expected %d)",
+                 index, rb_enc_name(enc), ENC_TO_ENCINDEX(enc));
     }
     if (rb_enc_autoload_p(enc) && rb_enc_autoload(enc) == -1) {
-	rb_loaderror("failed to load encoding (%s)",
-		     rb_enc_name(enc));
+        rb_loaderror("failed to load encoding (%s)",
+                     rb_enc_name(enc));
     }
     return enc;
 }
@@ -271,16 +265,16 @@ rb_to_encoding_index(VALUE enc)
 
     idx = enc_check_encoding(enc);
     if (idx >= 0) {
-	return idx;
+        return idx;
     }
     else if (NIL_P(enc = rb_check_string_type(enc))) {
-	return -1;
+        return -1;
     }
     if (!rb_enc_asciicompat(rb_enc_get(enc))) {
-	return -1;
+        return -1;
     }
     if (!(name = rb_str_to_cstr(enc))) {
-	return -1;
+        return -1;
     }
     return rb_enc_find_index(name);
 }
@@ -292,10 +286,10 @@ name_for_encoding(volatile VALUE *enc)
     const char *n;
 
     if (!rb_enc_asciicompat(rb_enc_get(name))) {
-	rb_raise(rb_eArgError, "invalid encoding name (non ASCII)");
+        rb_raise(rb_eArgError, "invalid encoding name (non ASCII)");
     }
     if (!(n = rb_str_to_cstr(name))) {
-	rb_raise(rb_eArgError, "invalid encoding name (NUL byte)");
+        rb_raise(rb_eArgError, "invalid encoding name (NUL byte)");
     }
     return n;
 }
@@ -314,7 +308,7 @@ str_to_encindex(VALUE enc)
 {
     int idx = str_find_encindex(enc);
     if (idx < 0) {
-	rb_raise(rb_eArgError, "unknown encoding name - %"PRIsVALUE, enc);
+        rb_raise(rb_eArgError, "unknown encoding name - %"PRIsVALUE, enc);
     }
     return idx;
 }
@@ -345,16 +339,10 @@ rb_find_encoding(VALUE enc)
 static int
 enc_table_expand(struct enc_table *enc_table, int newsize)
 {
-    struct rb_encoding_entry *ent;
-    int count = newsize;
-
-    if (enc_table->size >= newsize) return newsize;
-    newsize = (newsize + 7) / 8 * 8;
-    ent = REALLOC_N(enc_table->list, struct rb_encoding_entry, newsize);
-    memset(ent + enc_table->size, 0, sizeof(*ent)*(newsize - enc_table->size));
-    enc_table->list = ent;
-    enc_table->size = newsize;
-    return count;
+    if (newsize > ENCODING_LIST_CAPA) {
+        rb_raise(rb_eEncodingError, "too many encoding (> %d)", ENCODING_LIST_CAPA);
+    }
+    return newsize;
 }
 
 static int
@@ -365,20 +353,20 @@ enc_register_at(struct enc_table *enc_table, int index, const char *name, rb_enc
 
     if (!valid_encoding_name_p(name)) return -1;
     if (!ent->name) {
-	ent->name = name = strdup(name);
+        ent->name = name = strdup(name);
     }
     else if (STRCASECMP(name, ent->name)) {
-	return -1;
+        return -1;
     }
     encoding = (rb_raw_encoding *)ent->enc;
     if (!encoding) {
-	encoding = xmalloc(sizeof(rb_encoding));
+        encoding = xmalloc(sizeof(rb_encoding));
     }
     if (base_encoding) {
-	*encoding = *base_encoding;
+        *encoding = *base_encoding;
     }
     else {
-	memset(encoding, 0, sizeof(*ent->enc));
+        memset(encoding, 0, sizeof(*ent->enc));
     }
     encoding->name = name;
     encoding->ruby_encoding_index = index;
@@ -405,7 +393,7 @@ static rb_encoding *
 enc_from_index(struct enc_table *enc_table, int index)
 {
     if (UNLIKELY(index < 0 || enc_table->count <= (index &= ENC_INDEX_MASK))) {
-	return 0;
+        return 0;
     }
     return enc_table->list[index].enc;
 }
@@ -413,17 +401,7 @@ enc_from_index(struct enc_table *enc_table, int index)
 rb_encoding *
 rb_enc_from_index(int index)
 {
-    rb_encoding *enc;
-
-    switch (index) {
-      case ENCINDEX_ASCII:    return global_enc_ascii;
-      case ENCINDEX_UTF_8:    return global_enc_utf_8;
-      case ENCINDEX_US_ASCII: return global_enc_us_ascii;
-      default:
-        GLOBAL_ENC_TABLE_EVAL(enc_table,
-                              enc = enc_from_index(enc_table, index));
-        return enc;
-    }
+    return enc_from_index(&global_enc_table, index);
 }
 
 int
@@ -462,9 +440,9 @@ enc_registered(struct enc_table *enc_table, const char *name)
     st_data_t idx = 0;
 
     if (!name) return -1;
-    if (!enc_table->list) return -1;
+    if (!enc_table->names) return -1;
     if (st_lookup(enc_table->names, (st_data_t)name, &idx)) {
-	return (int)idx;
+        return (int)idx;
     }
     return -1;
 }
@@ -484,10 +462,13 @@ rb_encdb_declare(const char *name)
 }
 
 static void
-enc_check_duplication(struct enc_table *enc_table, const char *name)
+enc_check_addable(struct enc_table *enc_table, const char *name)
 {
     if (enc_registered(enc_table, name) >= 0) {
-	rb_raise(rb_eArgError, "encoding %s is already registered", name);
+        rb_raise(rb_eArgError, "encoding %s is already registered", name);
+    }
+    else if (!valid_encoding_name_p(name)) {
+        rb_raise(rb_eArgError, "invalid encoding name: %s", name);
     }
 }
 
@@ -524,11 +505,7 @@ rb_enc_set_base(const char *name, const char *orig)
 int
 rb_enc_set_dummy(int index)
 {
-    rb_encoding *enc;
-
-    GLOBAL_ENC_TABLE_EVAL(enc_table,
-                          enc = enc_table->list[index].enc);
-
+    rb_encoding *enc = global_enc_table.list[index].enc;
     ENC_SET_DUMMY((rb_raw_encoding *)enc);
     return index;
 }
@@ -538,7 +515,7 @@ enc_replicate(struct enc_table *enc_table, const char *name, rb_encoding *encodi
 {
     int idx;
 
-    enc_check_duplication(enc_table, name);
+    enc_check_addable(enc_table, name);
     idx = enc_register(enc_table, name, encoding);
     if (idx < 0) rb_raise(rb_eArgError, "invalid encoding name: %s", name);
     set_base_encoding(enc_table, idx, encoding);
@@ -546,34 +523,6 @@ enc_replicate(struct enc_table *enc_table, const char *name, rb_encoding *encodi
     return idx;
 }
 
-int
-rb_enc_replicate(const char *name, rb_encoding *encoding)
-{
-    int r;
-
-    GLOBAL_ENC_TABLE_EVAL(enc_table,
-                          r = enc_replicate(enc_table, name, encoding));
-
-    return r;
-}
-
-/*
- * call-seq:
- *   enc.replicate(name) -> encoding
- *
- * Returns a replicated encoding of _enc_ whose name is _name_.
- * The new encoding should have the same byte structure of _enc_.
- * If _name_ is used by another encoding, raise ArgumentError.
- *
- */
-static VALUE
-enc_replicate_m(VALUE encoding, VALUE name)
-{
-    int idx = rb_enc_replicate(name_for_encoding(&name), rb_to_encoding(encoding));
-    RB_GC_GUARD(name);
-    return rb_enc_from_encoding_index(idx);
-}
-
 static int
 enc_replicate_with_index(struct enc_table *enc_table, const char *name, rb_encoding *origenc, int idx)
 {
@@ -581,11 +530,11 @@ enc_replicate_with_index(struct enc_table *enc_table, const char *name, rb_encod
         idx = enc_register(enc_table, name, origenc);
     }
     else {
-	idx = enc_register_at(enc_table, idx, name, origenc);
+        idx = enc_register_at(enc_table, idx, name, origenc);
     }
     if (idx >= 0) {
-	set_base_encoding(enc_table, idx, origenc);
-	set_encoding_const(name, rb_enc_from_index(idx));
+        set_base_encoding(enc_table, idx, origenc);
+        set_encoding_const(name, rb_enc_from_index(idx));
     }
     else {
         rb_raise(rb_eArgError, "failed to replicate encoding");
@@ -705,7 +654,7 @@ static int
 enc_alias_internal(struct enc_table *enc_table, const char *alias, int idx)
 {
     return st_insert2(enc_table->names, (st_data_t)alias, (st_data_t)idx,
-		      enc_dup_name);
+                      enc_dup_name);
 }
 
 static int
@@ -713,7 +662,7 @@ enc_alias(struct enc_table *enc_table, const char *alias, int idx)
 {
     if (!valid_encoding_name_p(alias)) return -1;
     if (!enc_alias_internal(enc_table, alias, idx))
-	set_encoding_const(alias, enc_from_index(enc_table, idx));
+        set_encoding_const(alias, enc_from_index(enc_table, idx));
     return idx;
 }
 
@@ -724,7 +673,7 @@ rb_enc_alias(const char *alias, const char *orig)
 
     GLOBAL_ENC_TABLE_ENTER(enc_table);
     {
-        enc_check_duplication(enc_table, alias);
+        enc_check_addable(enc_table, alias);
         if ((idx = rb_enc_find_index(orig)) < 0) {
             r =  -1;
         }
@@ -756,29 +705,23 @@ rb_encdb_alias(const char *alias, const char *orig)
     return r;
 }
 
-void
-rb_encdb_set_unicode(int index)
-{
-    rb_raw_encoding *enc = (rb_raw_encoding *)rb_enc_from_index(index);
-    ASSUME(enc);
-    enc->flags |= ONIGENC_FLAG_UNICODE;
-}
-
 static void
 rb_enc_init(struct enc_table *enc_table)
 {
     enc_table_expand(enc_table, ENCODING_COUNT + 1);
     if (!enc_table->names) {
-	enc_table->names = st_init_strcasetable();
+        enc_table->names = st_init_strcasetable_with_size(ENCODING_LIST_CAPA);
     }
+#define OnigEncodingASCII_8BIT OnigEncodingASCII
 #define ENC_REGISTER(enc) enc_register_at(enc_table, ENCINDEX_##enc, rb_enc_name(&OnigEncoding##enc), &OnigEncoding##enc)
-    ENC_REGISTER(ASCII);
+    ENC_REGISTER(ASCII_8BIT);
     ENC_REGISTER(UTF_8);
     ENC_REGISTER(US_ASCII);
-    global_enc_ascii = enc_table->list[ENCINDEX_ASCII].enc;
+    global_enc_ascii = enc_table->list[ENCINDEX_ASCII_8BIT].enc;
     global_enc_utf_8 = enc_table->list[ENCINDEX_UTF_8].enc;
     global_enc_us_ascii = enc_table->list[ENCINDEX_US_ASCII].enc;
 #undef ENC_REGISTER
+#undef OnigEncodingASCII_8BIT
 #define ENCDB_REGISTER(name, enc) enc_register_at(enc_table, ENCINDEX_##enc, name, NULL)
     ENCDB_REGISTER("UTF-16BE", UTF_16BE);
     ENCDB_REGISTER("UTF-16LE", UTF_16LE);
@@ -813,9 +756,9 @@ load_encoding(const char *name)
     int idx;
 
     while (s < e) {
-	if (!ISALNUM(*s)) *s = '_';
-	else if (ISUPPER(*s)) *s = (char)TOLOWER(*s);
-	++s;
+        if (!ISALNUM(*s)) *s = '_';
+        else if (ISUPPER(*s)) *s = (char)TOLOWER(*s);
+        ++s;
     }
     enclib = rb_fstring(enclib);
     ruby_debug = Qfalse;
@@ -848,16 +791,16 @@ enc_autoload_body(struct enc_table *enc_table, rb_encoding *enc)
 
     if (base) {
         int i = 0;
-	do {
-	    if (i >= enc_table->count) return -1;
-	} while (enc_table->list[i].enc != base && (++i, 1));
-	if (rb_enc_autoload_p(base)) {
-	    if (rb_enc_autoload(base) < 0) return -1;
-	}
-	i = enc->ruby_encoding_index;
-	enc_register_at(enc_table, i & ENC_INDEX_MASK, rb_enc_name(enc), base);
+        do {
+            if (i >= enc_table->count) return -1;
+        } while (enc_table->list[i].enc != base && (++i, 1));
+        if (rb_enc_autoload_p(base)) {
+            if (rb_enc_autoload(base) < 0) return -1;
+        }
+        i = enc->ruby_encoding_index;
+        enc_register_at(enc_table, i & ENC_INDEX_MASK, rb_enc_name(enc), base);
         ((rb_raw_encoding *)enc)->ruby_encoding_index = i;
-	i &= ENC_INDEX_MASK;
+        i &= ENC_INDEX_MASK;
         return i;
     }
     else {
@@ -880,25 +823,23 @@ rb_enc_autoload(rb_encoding *enc)
 int
 rb_enc_find_index(const char *name)
 {
-    int i;
+    int i = enc_registered(&global_enc_table, name);
     rb_encoding *enc;
 
-    GLOBAL_ENC_TABLE_EVAL(enc_table, i = enc_registered(enc_table, name));
-
     if (i < 0) {
-	i = load_encoding(name);
+        i = load_encoding(name);
     }
     else if (!(enc = rb_enc_from_index(i))) {
-	if (i != UNSPECIFIED_ENCODING) {
-	    rb_raise(rb_eArgError, "encoding %s is not registered", name);
-	}
+        if (i != UNSPECIFIED_ENCODING) {
+            rb_raise(rb_eArgError, "encoding %s is not registered", name);
+        }
     }
     else if (rb_enc_autoload_p(enc)) {
-	if (rb_enc_autoload(enc) < 0) {
-	    rb_warn("failed to load encoding (%s); use ASCII-8BIT instead",
-		    name);
-	    return 0;
-	}
+        if (rb_enc_autoload(enc) < 0) {
+            rb_warn("failed to load encoding (%s); use ASCII-8BIT instead",
+                    name);
+            return 0;
+        }
     }
     return i;
 }
@@ -931,11 +872,11 @@ enc_capable(VALUE obj)
       case T_REGEXP:
       case T_FILE:
       case T_SYMBOL:
-	return TRUE;
+        return TRUE;
       case T_DATA:
-	if (is_data_encoding(obj)) return TRUE;
+        if (is_data_encoding(obj)) return TRUE;
       default:
-	return FALSE;
+        return FALSE;
     }
 }
 
@@ -957,11 +898,11 @@ enc_get_index_str(VALUE str)
 {
     int i = ENCODING_GET_INLINED(str);
     if (i == ENCODING_INLINE_MAX) {
-	VALUE iv;
+        VALUE iv;
 
 #if 0
-	iv = rb_ivar_get(str, rb_id_encoding());
-	i = NUM2INT(iv);
+        iv = rb_ivar_get(str, rb_id_encoding());
+        i = NUM2INT(iv);
 #else
         /*
          * Tentatively, assume ASCII-8BIT, if encoding index instance
@@ -969,7 +910,7 @@ enc_get_index_str(VALUE str)
          * all instance variables are removed in `obj_free`.
          */
         iv = rb_attr_get(str, rb_id_encoding());
-        i = NIL_P(iv) ? ENCINDEX_ASCII : NUM2INT(iv);
+        i = NIL_P(iv) ? ENCINDEX_ASCII_8BIT : NUM2INT(iv);
 #endif
     }
     return i;
@@ -982,31 +923,31 @@ rb_enc_get_index(VALUE obj)
     VALUE tmp;
 
     if (SPECIAL_CONST_P(obj)) {
-	if (!SYMBOL_P(obj)) return -1;
-	obj = rb_sym2str(obj);
+        if (!SYMBOL_P(obj)) return -1;
+        obj = rb_sym2str(obj);
     }
     switch (BUILTIN_TYPE(obj)) {
       case T_STRING:
       case T_SYMBOL:
       case T_REGEXP:
-	i = enc_get_index_str(obj);
-	break;
+        i = enc_get_index_str(obj);
+        break;
       case T_FILE:
-	tmp = rb_funcallv(obj, rb_intern("internal_encoding"), 0, 0);
-	if (NIL_P(tmp)) {
-	    tmp = rb_funcallv(obj, rb_intern("external_encoding"), 0, 0);
-	}
-	if (is_obj_encoding(tmp)) {
-	    i = enc_check_encoding(tmp);
-	}
-	break;
+        tmp = rb_funcallv(obj, rb_intern("internal_encoding"), 0, 0);
+        if (NIL_P(tmp)) {
+            tmp = rb_funcallv(obj, rb_intern("external_encoding"), 0, 0);
+        }
+        if (is_obj_encoding(tmp)) {
+            i = enc_check_encoding(tmp);
+        }
+        break;
       case T_DATA:
-	if (is_data_encoding(obj)) {
-	    i = enc_check_encoding(obj);
-	}
-	break;
+        if (is_data_encoding(obj)) {
+            i = enc_check_encoding(obj);
+        }
+        break;
       default:
-	break;
+        break;
     }
     return i;
 }
@@ -1019,8 +960,8 @@ enc_set_index(VALUE obj, int idx)
     }
 
     if (idx < ENCODING_INLINE_MAX) {
-	ENCODING_SET_INLINED(obj, idx);
-	return;
+        ENCODING_SET_INLINED(obj, idx);
+        return;
     }
     ENCODING_SET_INLINED(obj, ENCODING_INLINE_MAX);
     rb_ivar_set(obj, rb_id_encoding(), INT2NUM(idx));
@@ -1044,19 +985,19 @@ rb_enc_associate_index(VALUE obj, int idx)
     rb_check_frozen(obj);
     oldidx = rb_enc_get_index(obj);
     if (oldidx == idx)
-	return obj;
+        return obj;
     if (SPECIAL_CONST_P(obj)) {
-	rb_raise(rb_eArgError, "cannot set encoding");
+        rb_raise(rb_eArgError, "cannot set encoding");
     }
     enc = must_encindex(idx);
     if (!ENC_CODERANGE_ASCIIONLY(obj) ||
-	!rb_enc_asciicompat(enc)) {
-	ENC_CODERANGE_CLEAR(obj);
+        !rb_enc_asciicompat(enc)) {
+        ENC_CODERANGE_CLEAR(obj);
     }
     termlen = rb_enc_mbminlen(enc);
     oldtermlen = rb_enc_mbminlen(rb_enc_from_index(oldidx));
     if (oldtermlen != termlen && RB_TYPE_P(obj, T_STRING)) {
-	rb_str_change_terminator_length(obj, oldtermlen, termlen);
+        rb_str_change_terminator_length(obj, oldtermlen, termlen);
     }
     enc_set_index(obj, idx);
     return obj;
@@ -1074,13 +1015,22 @@ rb_enc_get(VALUE obj)
     return rb_enc_from_index(rb_enc_get_index(obj));
 }
 
+const char *
+rb_enc_inspect_name(rb_encoding *enc)
+{
+    if (enc == global_enc_ascii) {
+        return "BINARY (ASCII-8BIT)";
+    }
+    return enc->name;
+}
+
 static rb_encoding*
 rb_encoding_check(rb_encoding* enc, VALUE str1, VALUE str2)
 {
     if (!enc)
-	rb_raise(rb_eEncCompatError, "incompatible character encodings: %s and %s",
-		 rb_enc_name(rb_enc_get(str1)),
-		 rb_enc_name(rb_enc_get(str2)));
+        rb_raise(rb_eEncCompatError, "incompatible character encodings: %s and %s",
+                 rb_enc_inspect_name(rb_enc_get(str1)),
+                 rb_enc_inspect_name(rb_enc_get(str2)));
     return enc;
 }
 
@@ -1109,48 +1059,48 @@ enc_compatible_latter(VALUE str1, VALUE str2, int idx1, int idx2)
 
     isstr2 = RB_TYPE_P(str2, T_STRING);
     if (isstr2 && RSTRING_LEN(str2) == 0)
-	return enc1;
+        return enc1;
     isstr1 = RB_TYPE_P(str1, T_STRING);
     if (isstr1 && isstr2 && RSTRING_LEN(str1) == 0)
-	return (rb_enc_asciicompat(enc1) && rb_enc_str_asciionly_p(str2)) ? enc1 : enc2;
+        return (rb_enc_asciicompat(enc1) && rb_enc_str_asciionly_p(str2)) ? enc1 : enc2;
     if (!rb_enc_asciicompat(enc1) || !rb_enc_asciicompat(enc2)) {
-	return 0;
+        return 0;
     }
 
     /* objects whose encoding is the same of contents */
     if (!isstr2 && idx2 == ENCINDEX_US_ASCII)
-	return enc1;
+        return enc1;
     if (!isstr1 && idx1 == ENCINDEX_US_ASCII)
-	return enc2;
+        return enc2;
 
     if (!isstr1) {
-	VALUE tmp = str1;
-	int idx0 = idx1;
-	str1 = str2;
-	str2 = tmp;
-	idx1 = idx2;
-	idx2 = idx0;
-	idx0 = isstr1;
-	isstr1 = isstr2;
-	isstr2 = idx0;
+        VALUE tmp = str1;
+        int idx0 = idx1;
+        str1 = str2;
+        str2 = tmp;
+        idx1 = idx2;
+        idx2 = idx0;
+        idx0 = isstr1;
+        isstr1 = isstr2;
+        isstr2 = idx0;
     }
     if (isstr1) {
-	int cr1, cr2;
+        int cr1, cr2;
 
-	cr1 = rb_enc_str_coderange(str1);
-	if (isstr2) {
-	    cr2 = rb_enc_str_coderange(str2);
-	    if (cr1 != cr2) {
-		/* may need to handle ENC_CODERANGE_BROKEN */
-		if (cr1 == ENC_CODERANGE_7BIT) return enc2;
-		if (cr2 == ENC_CODERANGE_7BIT) return enc1;
-	    }
-	    if (cr2 == ENC_CODERANGE_7BIT) {
-		return enc1;
-	    }
-	}
-	if (cr1 == ENC_CODERANGE_7BIT)
-	    return enc2;
+        cr1 = rb_enc_str_coderange(str1);
+        if (isstr2) {
+            cr2 = rb_enc_str_coderange(str2);
+            if (cr1 != cr2) {
+                /* may need to handle ENC_CODERANGE_BROKEN */
+                if (cr1 == ENC_CODERANGE_7BIT) return enc2;
+                if (cr2 == ENC_CODERANGE_7BIT) return enc1;
+            }
+            if (cr2 == ENC_CODERANGE_7BIT) {
+                return enc1;
+            }
+        }
+        if (cr1 == ENC_CODERANGE_7BIT)
+            return enc2;
     }
     return 0;
 }
@@ -1165,10 +1115,10 @@ enc_compatible_str(VALUE str1, VALUE str2)
         return 0;
 
     if (idx1 == idx2) {
-	return rb_enc_from_index(idx1);
+        return rb_enc_from_index(idx1);
     }
     else {
-	return enc_compatible_latter(str1, str2, idx1, idx2);
+        return enc_compatible_latter(str1, str2, idx1, idx2);
     }
 }
 
@@ -1182,7 +1132,7 @@ rb_enc_compatible(VALUE str1, VALUE str2)
         return 0;
 
     if (idx1 == idx2) {
-	return rb_enc_from_index(idx1);
+        return rb_enc_from_index(idx1);
     }
 
     return enc_compatible_latter(str1, str2, idx1, idx2);
@@ -1207,7 +1157,7 @@ rb_obj_encoding(VALUE obj)
 {
     int idx = rb_enc_get_index(obj);
     if (idx < 0) {
-	rb_raise(rb_eTypeError, "unknown encoding");
+        rb_raise(rb_eTypeError, "unknown encoding");
     }
     return rb_enc_from_encoding_index(idx & ENC_INDEX_MASK);
 }
@@ -1274,7 +1224,7 @@ rb_enc_codepoint_len(const char *p, const char *e, int *len_p, rb_encoding *enc)
         rb_raise(rb_eArgError, "empty string");
     r = rb_enc_precise_mbclen(p, e, enc);
     if (!MBCLEN_CHARFOUND_P(r)) {
-	rb_raise(rb_eArgError, "invalid byte sequence in %s", rb_enc_name(enc));
+        rb_raise(rb_eArgError, "invalid byte sequence in %s", rb_enc_name(enc));
     }
     if (len_p) *len_p = MBCLEN_CHARFOUND_LEN(r);
     return rb_enc_mbc_to_codepoint(p, e, enc);
@@ -1285,7 +1235,7 @@ rb_enc_codelen(int c, rb_encoding *enc)
 {
     int n = ONIGENC_CODE_TO_MBCLEN(enc,c);
     if (n == 0) {
-	rb_raise(rb_eArgError, "invalid codepoint 0x%x in %s", c, rb_enc_name(enc));
+        rb_raise(rb_eArgError, "invalid codepoint 0x%x in %s", c, rb_enc_name(enc));
     }
     return n;
 }
@@ -1317,16 +1267,17 @@ enc_inspect(VALUE self)
     rb_encoding *enc;
 
     if (!is_data_encoding(self)) {
-	not_encoding(self);
+        not_encoding(self);
     }
     if (!(enc = DATA_PTR(self)) || rb_enc_from_index(rb_enc_to_index(enc)) != enc) {
-	rb_raise(rb_eTypeError, "broken Encoding");
+        rb_raise(rb_eTypeError, "broken Encoding");
     }
+
     return rb_enc_sprintf(rb_usascii_encoding(),
-			  "#<%"PRIsVALUE":%s%s%s>", rb_obj_class(self),
-			  rb_enc_name(enc),
-			  (ENC_DUMMY_P(enc) ? " (dummy)" : ""),
-			  rb_enc_autoload_p(enc) ? " (autoload)" : "");
+                          "#<%"PRIsVALUE":%s%s%s>", rb_obj_class(self),
+                          rb_enc_inspect_name(enc),
+                          (ENC_DUMMY_P(enc) ? " (dummy)" : ""),
+                          rb_enc_autoload_p(enc) ? " (autoload)" : "");
 }
 
 /*
@@ -1350,8 +1301,8 @@ enc_names_i(st_data_t name, st_data_t idx, st_data_t args)
     VALUE *arg = (VALUE *)args;
 
     if ((int)idx == (int)arg[0]) {
-	VALUE str = rb_fstring_cstr((char *)name);
-	rb_ary_push(arg[1], str);
+        VALUE str = rb_fstring_cstr((char *)name);
+        rb_ary_push(arg[1], str);
     }
     return ST_CONTINUE;
 }
@@ -1371,10 +1322,7 @@ enc_names(VALUE self)
 
     args[0] = (VALUE)rb_to_encoding_index(self);
     args[1] = rb_ary_new2(0);
-
-    GLOBAL_ENC_TABLE_EVAL(enc_table,
-                          st_foreach(enc_table->names, enc_names_i, (st_data_t)args));
-
+    st_foreach(global_enc_table.names, enc_names_i, (st_data_t)args);
     return args[1];
 }
 
@@ -1400,14 +1348,7 @@ static VALUE
 enc_list(VALUE klass)
 {
     VALUE ary = rb_ary_new2(0);
-
-    RB_VM_LOCK_ENTER();
-    {
-        rb_ary_replace(ary, rb_default_encoding_list);
-        rb_ary_concat(ary, rb_additional_encoding_list);
-    }
-    RB_VM_LOCK_LEAVE();
-
+    rb_ary_replace(ary, rb_encoding_list);
     return ary;
 }
 
@@ -1438,7 +1379,7 @@ enc_find(VALUE klass, VALUE enc)
 {
     int idx;
     if (is_obj_encoding(enc))
-	return enc;
+        return enc;
     idx = str_to_encindex(enc);
     if (idx == UNSPECIFIED_ENCODING) return Qnil;
     return rb_enc_from_encoding_index(idx);
@@ -1520,7 +1461,7 @@ rb_ascii8bit_encoding(void)
 int
 rb_ascii8bit_encindex(void)
 {
-    return ENCINDEX_ASCII;
+    return ENCINDEX_ASCII_8BIT;
 }
 
 rb_encoding *
@@ -1556,15 +1497,17 @@ rb_locale_encindex(void)
 
     if (idx < 0) idx = ENCINDEX_UTF_8;
 
-    GLOBAL_ENC_TABLE_ENTER(enc_table);
-    if (enc_registered(enc_table, "locale") < 0) {
+    if (enc_registered(&global_enc_table, "locale") < 0) {
 # if defined _WIN32
-	void Init_w32_codepage(void);
-	Init_w32_codepage();
+        void Init_w32_codepage(void);
+        Init_w32_codepage();
 # endif
-	enc_alias_internal(enc_table, "locale", idx);
+        GLOBAL_ENC_TABLE_ENTER(enc_table);
+        {
+            enc_alias_internal(enc_table, "locale", idx);
+        }
+        GLOBAL_ENC_TABLE_LEAVE();
     }
-    GLOBAL_ENC_TABLE_LEAVE();
 
     return idx;
 }
@@ -1578,13 +1521,8 @@ rb_locale_encoding(void)
 int
 rb_filesystem_encindex(void)
 {
-    int idx;
-
-    GLOBAL_ENC_TABLE_EVAL(enc_table,
-                          idx = enc_registered(enc_table, "filesystem"));
-
-    if (idx < 0)
-	idx = ENCINDEX_ASCII;
+    int idx = enc_registered(&global_enc_table, "filesystem");
+    if (idx < 0) idx = ENCINDEX_ASCII_8BIT;
     return idx;
 }
 
@@ -1607,15 +1545,22 @@ enc_set_default_encoding(struct default_encoding *def, VALUE encoding, const cha
     int overridden = FALSE;
 
     if (def->index != -2)
-	/* Already set */
-	overridden = TRUE;
+        /* Already set */
+        overridden = TRUE;
 
     GLOBAL_ENC_TABLE_ENTER(enc_table);
     {
         if (NIL_P(encoding)) {
             def->index = -1;
             def->enc = 0;
-            st_insert(enc_table->names, (st_data_t)strdup(name),
+            char *name_dup = strdup(name);
+
+            st_data_t existing_name = (st_data_t)name_dup;
+            if (st_delete(enc_table->names, &existing_name, NULL)) {
+                xfree((void *)existing_name);
+            }
+
+            st_insert(enc_table->names, (st_data_t)name_dup,
                       (st_data_t)UNSPECIFIED_ENCODING);
         }
         else {
@@ -1807,45 +1752,45 @@ set_encoding_const(const char *name, rb_encoding *enc)
 
     if (ISDIGIT(*s)) return;
     if (ISUPPER(*s)) {
-	hasupper = 1;
-	while (*++s && (ISALNUM(*s) || *s == '_')) {
-	    if (ISLOWER(*s)) haslower = 1;
-	}
+        hasupper = 1;
+        while (*++s && (ISALNUM(*s) || *s == '_')) {
+            if (ISLOWER(*s)) haslower = 1;
+        }
     }
     if (!*s) {
-	if (s - name > ENCODING_NAMELEN_MAX) return;
-	valid = 1;
-	rb_define_const(rb_cEncoding, name, encoding);
+        if (s - name > ENCODING_NAMELEN_MAX) return;
+        valid = 1;
+        rb_define_const(rb_cEncoding, name, encoding);
     }
     if (!valid || haslower) {
-	size_t len = s - name;
-	if (len > ENCODING_NAMELEN_MAX) return;
-	if (!haslower || !hasupper) {
-	    do {
-		if (ISLOWER(*s)) haslower = 1;
-		if (ISUPPER(*s)) hasupper = 1;
-	    } while (*++s && (!haslower || !hasupper));
-	    len = s - name;
-	}
-	len += strlen(s);
-	if (len++ > ENCODING_NAMELEN_MAX) return;
-	MEMCPY(s = ALLOCA_N(char, len), name, char, len);
-	name = s;
-	if (!valid) {
-	    if (ISLOWER(*s)) *s = ONIGENC_ASCII_CODE_TO_UPPER_CASE((int)*s);
-	    for (; *s; ++s) {
-		if (!ISALNUM(*s)) *s = '_';
-	    }
-	    if (hasupper) {
-		rb_define_const(rb_cEncoding, name, encoding);
-	    }
-	}
-	if (haslower) {
-	    for (s = (char *)name; *s; ++s) {
-		if (ISLOWER(*s)) *s = ONIGENC_ASCII_CODE_TO_UPPER_CASE((int)*s);
-	    }
-	    rb_define_const(rb_cEncoding, name, encoding);
-	}
+        size_t len = s - name;
+        if (len > ENCODING_NAMELEN_MAX) return;
+        if (!haslower || !hasupper) {
+            do {
+                if (ISLOWER(*s)) haslower = 1;
+                if (ISUPPER(*s)) hasupper = 1;
+            } while (*++s && (!haslower || !hasupper));
+            len = s - name;
+        }
+        len += strlen(s);
+        if (len++ > ENCODING_NAMELEN_MAX) return;
+        MEMCPY(s = ALLOCA_N(char, len), name, char, len);
+        name = s;
+        if (!valid) {
+            if (ISLOWER(*s)) *s = ONIGENC_ASCII_CODE_TO_UPPER_CASE((int)*s);
+            for (; *s; ++s) {
+                if (!ISALNUM(*s)) *s = '_';
+            }
+            if (hasupper) {
+                rb_define_const(rb_cEncoding, name, encoding);
+            }
+        }
+        if (haslower) {
+            for (s = (char *)name; *s; ++s) {
+                if (ISLOWER(*s)) *s = ONIGENC_ASCII_CODE_TO_UPPER_CASE((int)*s);
+            }
+            rb_define_const(rb_cEncoding, name, encoding);
+        }
     }
 }
 
@@ -1875,15 +1820,8 @@ rb_enc_name_list_i(st_data_t name, st_data_t idx, st_data_t arg)
 static VALUE
 rb_enc_name_list(VALUE klass)
 {
-    VALUE ary;
-
-    GLOBAL_ENC_TABLE_ENTER(enc_table);
-    {
-        ary = rb_ary_new2(enc_table->names->num_entries);
-        st_foreach(enc_table->names, rb_enc_name_list_i, (st_data_t)ary);
-    }
-    GLOBAL_ENC_TABLE_LEAVE();
-
+    VALUE ary = rb_ary_new2(global_enc_table.names->num_entries);
+    st_foreach(global_enc_table.names, rb_enc_name_list_i, (st_data_t)ary);
     return ary;
 }
 
@@ -1896,14 +1834,14 @@ rb_enc_aliases_enc_i(st_data_t name, st_data_t orig, st_data_t arg)
     VALUE key, str = rb_ary_entry(ary, idx);
 
     if (NIL_P(str)) {
-	rb_encoding *enc = rb_enc_from_index(idx);
+        rb_encoding *enc = rb_enc_from_index(idx);
 
-	if (!enc) return ST_CONTINUE;
-	if (STRCASECMP((char*)name, rb_enc_name(enc)) == 0) {
-	    return ST_CONTINUE;
-	}
-	str = rb_fstring_cstr(rb_enc_name(enc));
-	rb_ary_store(ary, idx, str);
+        if (!enc) return ST_CONTINUE;
+        if (STRCASECMP((char*)name, rb_enc_name(enc)) == 0) {
+            return ST_CONTINUE;
+        }
+        str = rb_fstring_cstr(rb_enc_name(enc));
+        rb_ary_store(ary, idx, str);
     }
     key = rb_fstring_cstr((char *)name);
     rb_hash_aset(aliases, key, str);
@@ -1929,8 +1867,7 @@ rb_enc_aliases(VALUE klass)
     aliases[0] = rb_hash_new();
     aliases[1] = rb_ary_new();
 
-    GLOBAL_ENC_TABLE_EVAL(enc_table,
-                          st_foreach(enc_table->names, rb_enc_aliases_enc_i, (st_data_t)aliases));
+    st_foreach(global_enc_table.names, rb_enc_aliases_enc_i, (st_data_t)aliases);
 
     return aliases[0];
 }
@@ -1981,7 +1918,6 @@ Init_Encoding(void)
     rb_define_method(rb_cEncoding, "names", enc_names, 0);
     rb_define_method(rb_cEncoding, "dummy?", enc_dummy_p, 0);
     rb_define_method(rb_cEncoding, "ascii_compatible?", enc_ascii_compatible_p, 0);
-    rb_define_method(rb_cEncoding, "replicate", enc_replicate_m, 1);
     rb_define_singleton_method(rb_cEncoding, "list", enc_list, 0);
     rb_define_singleton_method(rb_cEncoding, "name_list", rb_enc_name_list, 0);
     rb_define_singleton_method(rb_cEncoding, "aliases", rb_enc_aliases, 0);
@@ -1999,18 +1935,12 @@ Init_Encoding(void)
 
     struct enc_table *enc_table = &global_enc_table;
 
-    if (DEFAULT_ENCODING_LIST_CAPA < enc_table->count) rb_bug("DEFAULT_ENCODING_LIST_CAPA is too small");
-
-    list = rb_additional_encoding_list = rb_ary_new();
+    list = rb_encoding_list = rb_ary_new2(ENCODING_LIST_CAPA);
     RBASIC_CLEAR_CLASS(list);
-    rb_gc_register_mark_object(list);
-
-    list = rb_default_encoding_list = rb_ary_new2(DEFAULT_ENCODING_LIST_CAPA);
-    RBASIC_CLEAR_CLASS(list);
-    rb_gc_register_mark_object(list);
+    rb_vm_register_global_object(list);
 
     for (i = 0; i < enc_table->count; ++i) {
-	rb_ary_push(list, enc_new(enc_table->list[i].enc));
+        rb_ary_push(list, enc_new(enc_table->list[i].enc));
     }
 
     rb_marshal_define_compat(rb_cEncoding, Qnil, 0, enc_m_loader);
@@ -2027,5 +1957,5 @@ Init_encodings(void)
 void
 rb_enc_foreach_name(int (*func)(st_data_t name, st_data_t idx, st_data_t arg), st_data_t arg)
 {
-    GLOBAL_ENC_TABLE_EVAL(enc_table, st_foreach(enc_table->names, func, arg));
+    st_foreach(global_enc_table.names, func, arg);
 }
