@@ -29,15 +29,6 @@ static VALUE rb_eUncaughtThrow;
 static ID id_result, id_tag, id_value;
 #define id_mesg idMesg
 
-typedef enum call_type {
-    CALL_PUBLIC,
-    CALL_FCALL,
-    CALL_VCALL,
-    CALL_PUBLIC_KW,
-    CALL_FCALL_KW,
-    CALL_TYPE_MAX
-} call_type;
-
 static VALUE send_internal(int argc, const VALUE *argv, VALUE recv, call_type scope);
 static VALUE vm_call0_body(rb_execution_context_t* ec, struct rb_calling_info *calling, const VALUE *argv);
 
@@ -96,7 +87,10 @@ vm_call0_cc(rb_execution_context_t *ec, VALUE recv, ID id, int argc, const VALUE
     }
 
     struct rb_calling_info calling = {
-        .ci = &VM_CI_ON_STACK(id, flags, argc, NULL),
+        .cd = &(struct rb_call_data) {
+            .ci = &VM_CI_ON_STACK(id, flags, argc, NULL),
+            .cc = NULL,
+        },
         .cc = cc,
         .block_handler = vm_passed_block_handler(ec),
         .recv = recv,
@@ -117,7 +111,7 @@ vm_call0_cme(rb_execution_context_t *ec, struct rb_calling_info *calling, const
 static VALUE
 vm_call0_super(rb_execution_context_t *ec, struct rb_calling_info *calling, const VALUE *argv, VALUE klass, enum method_missing_reason ex)
 {
-    ID mid = vm_ci_mid(calling->ci);
+    ID mid = vm_ci_mid(calling->cd->ci);
     klass = RCLASS_SUPER(klass);
 
     if (klass) {
@@ -136,7 +130,7 @@ vm_call0_super(rb_execution_context_t *ec, struct rb_calling_info *calling, cons
 static VALUE
 vm_call0_cfunc_with_frame(rb_execution_context_t* ec, struct rb_calling_info *calling, const VALUE *argv)
 {
-    const struct rb_callinfo *ci = calling->ci;
+    const struct rb_callinfo *ci = calling->cd->ci;
     VALUE val;
     const rb_callable_method_entry_t *me = vm_cc_cme(calling->cc);
     const rb_method_cfunc_t *cfunc = UNALIGNED_MEMBER_PTR(me->def, body.cfunc);
@@ -201,7 +195,7 @@ vm_call_check_arity(struct rb_calling_info *calling, int argc, const VALUE *argv
 static VALUE
 vm_call0_body(rb_execution_context_t *ec, struct rb_calling_info *calling, const VALUE *argv)
 {
-    const struct rb_callinfo *ci = calling->ci;
+    const struct rb_callinfo *ci = calling->cd->ci;
     const struct rb_callcache *cc = calling->cc;
     VALUE ret;
 
@@ -296,11 +290,15 @@ vm_call0_body(rb_execution_context_t *ec, struct rb_calling_info *calling, const
             }
           case OPTIMIZED_METHOD_TYPE_STRUCT_AREF:
             vm_call_check_arity(calling, 0, argv);
-            ret = vm_call_opt_struct_aref0(ec, calling);
+            VM_CALL_METHOD_ATTR(ret,
+                                vm_call_opt_struct_aref0(ec, calling),
+                                (void)0);
             goto success;
           case OPTIMIZED_METHOD_TYPE_STRUCT_ASET:
             vm_call_check_arity(calling, 1, argv);
-            ret = vm_call_opt_struct_aset0(ec, calling, argv[0]);
+            VM_CALL_METHOD_ATTR(ret,
+                                vm_call_opt_struct_aset0(ec, calling, argv[0]),
+                                (void)0);
             goto success;
           default:
             rb_bug("vm_call0: unsupported optimized method type (%d)", vm_cc_cme(cc)->def->body.optimized.type);
@@ -396,71 +394,53 @@ NORETURN(static void uncallable_object(VALUE recv, ID mid));
 static inline const rb_callable_method_entry_t *rb_search_method_entry(VALUE recv, ID mid);
 static inline enum method_missing_reason rb_method_call_status(rb_execution_context_t *ec, const rb_callable_method_entry_t *me, call_type scope, VALUE self);
 
-static const struct rb_callcache *
-cc_new(VALUE klass, ID mid, int argc, const rb_callable_method_entry_t *cme)
-{
-    const struct rb_callcache *cc = NULL;
-
-    RB_VM_LOCK_ENTER();
-    {
-        struct rb_class_cc_entries *ccs;
-        struct rb_id_table *cc_tbl = RCLASS_CC_TBL(klass);
-        VALUE ccs_data;
-
-        if (rb_id_table_lookup(cc_tbl, mid, &ccs_data)) {
-            // ok
-            ccs = (struct rb_class_cc_entries *)ccs_data;
-        }
-        else {
-            ccs = vm_ccs_create(klass, cc_tbl, mid, cme);
-        }
-
-        for (int i=0; i<ccs->len; i++) {
-            cc = ccs->entries[i].cc;
-            if (vm_cc_cme(cc) == cme) {
-                break;
-            }
-            cc = NULL;
-        }
-
-        if (cc == NULL) {
-            const struct rb_callinfo *ci = vm_ci_new(mid, 0, argc, NULL); // TODO: proper ci
-            cc = vm_cc_new(klass, cme, vm_call_general);
-            METHOD_ENTRY_CACHED_SET((struct rb_callable_method_entry_struct *)cme);
-            vm_ccs_push(klass, ccs, ci, cc);
-        }
-    }
-    RB_VM_LOCK_LEAVE();
-
-    return cc;
-}
-
 static VALUE
 gccct_hash(VALUE klass, ID mid)
 {
     return (klass >> 3) ^ (VALUE)mid;
 }
 
-NOINLINE(static const struct rb_callcache *gccct_method_search_slowpath(rb_vm_t *vm, VALUE klass, ID mid, int argc, unsigned int index));
+NOINLINE(static const struct rb_callcache *gccct_method_search_slowpath(rb_vm_t *vm, VALUE klass, unsigned int index, const struct rb_callinfo * ci));
 
 static const struct rb_callcache *
-gccct_method_search_slowpath(rb_vm_t *vm, VALUE klass, ID mid, int argc, unsigned int index)
+gccct_method_search_slowpath(rb_vm_t *vm, VALUE klass, unsigned int index, const struct rb_callinfo *ci)
 {
-    const rb_callable_method_entry_t *cme = rb_callable_method_entry(klass, mid);
-    const struct rb_callcache *cc;
+    struct rb_call_data cd = {
+            .ci = ci,
+            .cc = NULL
+    };
 
-    if (cme != NULL) {
-        cc = cc_new(klass, mid, argc, cme);
-    }
-    else {
-        cc = NULL;
-    }
+    vm_search_method_slowpath0(vm->self, &cd, klass);
 
-    return vm->global_cc_cache_table[index] = cc;
+    return vm->global_cc_cache_table[index] = cd.cc;
+}
+
+static void
+scope_to_ci(call_type scope, ID mid, int argc, struct rb_callinfo *ci)
+{
+    int flags = 0;
+
+    switch(scope) {
+      case CALL_PUBLIC:
+        break;
+      case CALL_FCALL:
+        flags |= VM_CALL_FCALL;
+        break;
+      case CALL_VCALL:
+        flags |= VM_CALL_VCALL;
+        break;
+      case CALL_PUBLIC_KW:
+        flags |= VM_CALL_KWARG;
+        break;
+      case CALL_FCALL_KW:
+        flags |= (VM_CALL_KWARG | VM_CALL_FCALL);
+        break;
+    }
+    *ci = VM_CI_ON_STACK(mid, flags, argc, NULL);
 }
 
 static inline const struct rb_callcache *
-gccct_method_search(rb_execution_context_t *ec, VALUE recv, ID mid, int argc)
+gccct_method_search(rb_execution_context_t *ec, VALUE recv, ID mid, const struct rb_callinfo *ci)
 {
     VALUE klass;
 
@@ -495,24 +475,24 @@ gccct_method_search(rb_execution_context_t *ec, VALUE recv, ID mid, int argc)
     }
 
     RB_DEBUG_COUNTER_INC(gccct_miss);
-    return gccct_method_search_slowpath(vm, klass, mid, argc, index);
+    return gccct_method_search_slowpath(vm, klass, index, ci);
 }
 
-/*!
- * \internal
+/**
+ * @internal
  * calls the specified method.
  *
  * This function is called by functions in rb_call* family.
- * \param ec     current execution context
- * \param recv   receiver of the method
- * \param mid    an ID that represents the name of the method
- * \param argc   the number of method arguments
- * \param argv   a pointer to an array of method arguments
- * \param scope
- * \param self   self in the caller. Qundef means no self is considered and
+ * @param ec     current execution context
+ * @param recv   receiver of the method
+ * @param mid    an ID that represents the name of the method
+ * @param argc   the number of method arguments
+ * @param argv   a pointer to an array of method arguments
+ * @param scope
+ * @param self   self in the caller. Qundef means no self is considered and
  *               protected methods cannot be called
  *
- * \note \a self is used in order to controlling access to protected methods.
+ * @note `self` is used in order to controlling access to protected methods.
  */
 static inline VALUE
 rb_call0(rb_execution_context_t *ec,
@@ -536,13 +516,16 @@ rb_call0(rb_execution_context_t *ec,
         break;
     }
 
-    const struct rb_callcache *cc = gccct_method_search(ec, recv, mid, argc);
+    struct rb_callinfo ci;
+    scope_to_ci(scope, mid, argc, &ci);
+
+    const struct rb_callcache *cc = gccct_method_search(ec, recv, mid, &ci);
 
     if (scope == CALL_PUBLIC) {
         RB_DEBUG_COUNTER_INC(call0_public);
 
         const rb_callable_method_entry_t *cc_cme = cc ? vm_cc_cme(cc) : NULL;
-        const rb_callable_method_entry_t *cme = callable_method_entry_refeinements0(CLASS_OF(recv), mid, NULL, true, cc_cme);
+        const rb_callable_method_entry_t *cme = callable_method_entry_refinements0(CLASS_OF(recv), mid, NULL, true, cc_cme);
         call_status = rb_method_call_status(ec, cme, scope, self);
 
         if (UNLIKELY(call_status != MISSING_NONE)) {
@@ -793,29 +776,29 @@ uncallable_object(VALUE recv, ID mid)
 
     if (SPECIAL_CONST_P(recv)) {
         rb_raise(rb_eNotImpError,
-                 "method `%"PRIsVALUE"' called on unexpected immediate object (%p)",
+                 "method '%"PRIsVALUE"' called on unexpected immediate object (%p)",
                  mname, (void *)recv);
     }
     else if ((flags = RBASIC(recv)->flags) == 0) {
         rb_raise(rb_eNotImpError,
-                 "method `%"PRIsVALUE"' called on terminated object (%p)",
+                 "method '%"PRIsVALUE"' called on terminated object (%p)",
                  mname, (void *)recv);
     }
     else if (!(typestr = rb_type_str(type = BUILTIN_TYPE(recv)))) {
         rb_raise(rb_eNotImpError,
-                 "method `%"PRIsVALUE"' called on broken T_?""?""?(0x%02x) object"
+                 "method '%"PRIsVALUE"' called on broken T_?""?""?(0x%02x) object"
                  " (%p flags=0x%"PRIxVALUE")",
                  mname, type, (void *)recv, flags);
     }
     else if (T_OBJECT <= type && type < T_NIL) {
         rb_raise(rb_eNotImpError,
-                 "method `%"PRIsVALUE"' called on hidden %s object"
+                 "method '%"PRIsVALUE"' called on hidden %s object"
                  " (%p flags=0x%"PRIxVALUE")",
                  mname, typestr, (void *)recv, flags);
     }
     else {
         rb_raise(rb_eNotImpError,
-                 "method `%"PRIsVALUE"' called on unexpected %s object"
+                 "method '%"PRIsVALUE"' called on unexpected %s object"
                  " (%p flags=0x%"PRIxVALUE")",
                  mname, typestr, (void *)recv, flags);
     }
@@ -874,16 +857,16 @@ rb_method_call_status(rb_execution_context_t *ec, const rb_callable_method_entry
 }
 
 
-/*!
- * \internal
+/**
+ * @internal
  * calls the specified method.
  *
  * This function is called by functions in rb_call* family.
- * \param recv   receiver
- * \param mid    an ID that represents the name of the method
- * \param argc   the number of method arguments
- * \param argv   a pointer to an array of method arguments
- * \param scope
+ * @param recv   receiver
+ * @param mid    an ID that represents the name of the method
+ * @param argc   the number of method arguments
+ * @param argv   a pointer to an array of method arguments
+ * @param scope
  */
 static inline VALUE
 rb_call(VALUE recv, ID mid, int argc, const VALUE *argv, call_type scope)
@@ -949,7 +932,7 @@ rb_make_no_method_exception(VALUE exc, VALUE format, VALUE obj,
     VALUE name = argv[0];
 
     if (!format) {
-        format = rb_fstring_lit("undefined method `%1$s' for %3$s%4$s");
+        format = rb_fstring_lit("undefined method '%1$s' for %3$s%4$s");
     }
     if (exc == rb_eNoMethodError) {
         VALUE args = rb_ary_new4(argc - 1, argv + 1);
@@ -979,17 +962,17 @@ raise_method_missing(rb_execution_context_t *ec, int argc, const VALUE *argv, VA
     stack_check(ec);
 
     if (last_call_status & MISSING_PRIVATE) {
-        format = rb_fstring_lit("private method `%1$s' called for %3$s%4$s");
+        format = rb_fstring_lit("private method '%1$s' called for %3$s%4$s");
     }
     else if (last_call_status & MISSING_PROTECTED) {
-        format = rb_fstring_lit("protected method `%1$s' called for %3$s%4$s");
+        format = rb_fstring_lit("protected method '%1$s' called for %3$s%4$s");
     }
     else if (last_call_status & MISSING_VCALL) {
-        format = rb_fstring_lit("undefined local variable or method `%1$s' for %3$s%4$s");
+        format = rb_fstring_lit("undefined local variable or method '%1$s' for %3$s%4$s");
         exc = rb_eNameError;
     }
     else if (last_call_status & MISSING_SUPER) {
-        format = rb_fstring_lit("super: no superclass method `%1$s' for %3$s%4$s");
+        format = rb_fstring_lit("super: no superclass method '%1$s' for %3$s%4$s");
     }
 
     {
@@ -1053,7 +1036,11 @@ static inline VALUE
 rb_funcallv_scope(VALUE recv, ID mid, int argc, const VALUE *argv, call_type scope)
 {
     rb_execution_context_t *ec = GET_EC();
-    const struct rb_callcache *cc = gccct_method_search(ec, recv, mid, argc);
+
+    struct rb_callinfo ci;
+    scope_to_ci(scope, mid, argc, &ci);
+
+    const struct rb_callcache *cc = gccct_method_search(ec, recv, mid, &ci);
     VALUE self = ec->cfp->self;
 
     if (LIKELY(cc) &&
@@ -1101,7 +1088,7 @@ rb_apply(VALUE recv, ID mid, VALUE args)
         return ret;
     }
     argv = ALLOCA_N(VALUE, argc);
-    MEMCPY(argv, RARRAY_CONST_PTR_TRANSIENT(args), VALUE, argc);
+    MEMCPY(argv, RARRAY_CONST_PTR(args), VALUE, argc);
 
     return rb_funcallv(recv, mid, argc, argv);
 }
@@ -1134,15 +1121,15 @@ rb_funcall(VALUE recv, ID mid, int n, ...)
     return rb_funcallv(recv, mid, n, argv);
 }
 
-/*!
+/**
  * Calls a method only if it is the basic method of `ancestor`
  * otherwise returns Qundef;
- * \param recv   receiver of the method
- * \param mid    an ID that represents the name of the method
- * \param ancestor the Class that defined the basic method
- * \param argc   the number of arguments
- * \param argv   pointer to an array of method arguments
- * \param kw_splat bool
+ * @param recv   receiver of the method
+ * @param mid    an ID that represents the name of the method
+ * @param ancestor the Class that defined the basic method
+ * @param argc   the number of arguments
+ * @param argv   pointer to an array of method arguments
+ * @param kw_splat bool
  */
 VALUE
 rb_check_funcall_basic_kw(VALUE recv, ID mid, VALUE ancestor, int argc, const VALUE *argv, int kw_splat)
@@ -1613,22 +1600,165 @@ rb_each(VALUE obj)
     return rb_call(obj, idEach, 0, 0, CALL_FCALL);
 }
 
-void rb_parser_warn_location(VALUE, int);
+static VALUE eval_default_path = Qfalse;
 
-static VALUE eval_default_path;
+#define EVAL_LOCATION_MARK "eval at "
+#define EVAL_LOCATION_MARK_LEN (int)rb_strlen_lit(EVAL_LOCATION_MARK)
+
+static VALUE
+get_eval_default_path(void)
+{
+    int location_lineno;
+    VALUE location_path = rb_source_location(&location_lineno);
+    if (!NIL_P(location_path)) {
+        return rb_fstring(rb_sprintf("("EVAL_LOCATION_MARK"%"PRIsVALUE":%d)",
+                                     location_path, location_lineno));
+    }
+
+    if (!eval_default_path) {
+        eval_default_path = rb_fstring_lit("(eval)");
+        rb_vm_register_global_object(eval_default_path);
+    }
+    return eval_default_path;
+}
 
 static const rb_iseq_t *
-eval_make_iseq(VALUE src, VALUE fname, int line, const rb_binding_t *bind,
+pm_eval_make_iseq(VALUE src, VALUE fname, int line,
+        const struct rb_block *base_block)
+{
+    const rb_iseq_t *const parent = vm_block_iseq(base_block);
+    const rb_iseq_t *iseq = parent;
+    VALUE name = rb_fstring_lit("<compiled>");
+    if (!fname) {
+        fname = rb_source_location(&line);
+    }
+
+    if (!UNDEF_P(fname)) {
+        if (!NIL_P(fname)) fname = rb_fstring(fname);
+    }
+    else {
+        fname = get_eval_default_path();
+    }
+
+    pm_parse_result_t result = { 0 };
+    pm_options_line_set(&result.options, line);
+
+    // Cout scopes, one for each parent iseq, plus one for our local scope
+    int scopes_count = 0;
+    do {
+        scopes_count++;
+    } while ((iseq = ISEQ_BODY(iseq)->parent_iseq) && (ISEQ_BODY(iseq)->type != ISEQ_TYPE_TOP));
+    pm_options_scopes_init(&result.options, scopes_count + 1);
+
+    // Walk over the scope tree, adding known locals at the correct depths. The
+    // scope array should be deepest -> shallowest. so lower indexes in the
+    // scopes array refer to root nodes on the tree, and higher indexes are the
+    // leaf nodes.
+    iseq = parent;
+    for (int scopes_index = 0; scopes_index < scopes_count; scopes_index++) {
+        int locals_count = ISEQ_BODY(iseq)->local_table_size;
+        pm_options_scope_t *options_scope = &result.options.scopes[scopes_count - scopes_index - 1];
+        pm_options_scope_init(options_scope, locals_count);
+
+        for (int local_index = 0; local_index < locals_count; local_index++) {
+            pm_string_t *scope_local = &options_scope->locals[local_index];
+            ID local = ISEQ_BODY(iseq)->local_table[local_index];
+
+            if (rb_is_local_id(local)) {
+                const char *name = rb_id2name(local);
+                size_t length = strlen(name);
+
+                // Explicitly skip numbered parameters. These should not be sent
+                // into the eval.
+                if (length == 2 && name[0] == '_' && name[1] >= '1' && name[1] <= '9') {
+                    continue;
+                }
+
+                pm_string_constant_init(scope_local, name, strlen(name));
+            }
+        }
+
+        iseq = ISEQ_BODY(iseq)->parent_iseq;
+    }
+
+    // Add our empty local scope at the very end of the array for our eval
+    // scope's locals.
+    pm_options_scope_init(&result.options.scopes[scopes_count], 0);
+    VALUE error = pm_parse_string(&result, src, fname);
+
+    // If the parse failed, clean up and raise.
+    if (error != Qnil) {
+        pm_parse_result_free(&result);
+        rb_exc_raise(error);
+    }
+
+    // Create one scope node for each scope passed in, initialize the local
+    // lookup table with all the local variable information attached to the
+    // scope used by the parser.
+    pm_scope_node_t *node = &result.node;
+    iseq = parent;
+
+    for (int scopes_index = 0; scopes_index < scopes_count; scopes_index++) {
+        pm_scope_node_t *parent_scope = ruby_xcalloc(1, sizeof(pm_scope_node_t));
+        RUBY_ASSERT(parent_scope != NULL);
+
+        pm_options_scope_t *options_scope = &result.options.scopes[scopes_count - scopes_index - 1];
+        parent_scope->parser = &result.parser;
+        parent_scope->index_lookup_table = st_init_numtable();
+
+        int locals_count = ISEQ_BODY(iseq)->local_table_size;
+        parent_scope->local_table_for_iseq_size = locals_count;
+        pm_constant_id_list_init(&parent_scope->locals);
+
+        for (int local_index = 0; local_index < locals_count; local_index++) {
+            const pm_string_t *scope_local = &options_scope->locals[local_index];
+
+            pm_constant_id_t constant_id = 0;
+            if (pm_string_length(scope_local) > 0) {
+                constant_id = pm_constant_pool_insert_constant(
+                        &result.parser.constant_pool, pm_string_source(scope_local),
+                        pm_string_length(scope_local));
+                st_insert(parent_scope->index_lookup_table, (st_data_t)constant_id, (st_data_t)local_index);
+            }
+            pm_constant_id_list_append(&parent_scope->locals, constant_id);
+        }
+
+        node->previous = parent_scope;
+        node = parent_scope;
+        iseq = ISEQ_BODY(iseq)->parent_iseq;
+    }
+
+    iseq = pm_iseq_new_eval(&result.node, name, fname, Qnil, line, parent, 0);
+
+    pm_scope_node_t *prev = result.node.previous;
+    while (prev) {
+        pm_scope_node_t *next = prev->previous;
+        ruby_xfree(prev);
+        prev = next;
+    }
+
+    pm_parse_result_free(&result);
+    rb_exec_event_hook_script_compiled(GET_EC(), iseq, src);
+
+    return iseq;
+}
+
+static const rb_iseq_t *
+eval_make_iseq(VALUE src, VALUE fname, int line,
                const struct rb_block *base_block)
 {
+    if (*rb_ruby_prism_ptr()) {
+        return pm_eval_make_iseq(src, fname, line, base_block);
+    }
     const VALUE parser = rb_parser_new();
     const rb_iseq_t *const parent = vm_block_iseq(base_block);
     rb_iseq_t *iseq = NULL;
+    VALUE ast_value;
     rb_ast_t *ast;
     int isolated_depth = 0;
 
     // Conditionally enable coverage depending on the current mode:
-    VALUE coverage_enabled = RBOOL(rb_get_coverage_mode() & COVERAGE_TARGET_EVAL);
+    int coverage_enabled = (rb_get_coverage_mode() & COVERAGE_TARGET_EVAL) != 0;
 
     {
         int depth = 1;
@@ -1655,24 +1785,19 @@ eval_make_iseq(VALUE src, VALUE fname, int line, const rb_binding_t *bind,
         if (!NIL_P(fname)) fname = rb_fstring(fname);
     }
     else {
-        fname = rb_fstring_lit("(eval)");
-        if (!eval_default_path) {
-            eval_default_path = rb_fstring_lit("(eval)");
-            rb_gc_register_mark_object(eval_default_path);
-        }
-        fname = eval_default_path;
-        coverage_enabled = Qfalse;
+        fname = get_eval_default_path();
+        coverage_enabled = FALSE;
     }
 
     rb_parser_set_context(parser, parent, FALSE);
-    ast = rb_parser_compile_string_path(parser, fname, src, line);
-    if (ast->body.root) {
-        if (ast->body.compile_option == Qnil) {
-            ast->body.compile_option = rb_obj_hide(rb_ident_hash_new());
-        }
-        rb_hash_aset(ast->body.compile_option, rb_sym_intern_ascii_cstr("coverage_enabled"), coverage_enabled);
+    if (ruby_vm_keep_script_lines) rb_parser_set_script_lines(parser);
+    ast_value = rb_parser_compile_string_path(parser, fname, src, line);
 
-        iseq = rb_iseq_new_eval(&ast->body,
+    ast = rb_ruby_ast_data_get(ast_value);
+
+    if (ast->body.root) {
+        ast->body.coverage_enabled = coverage_enabled;
+        iseq = rb_iseq_new_eval(ast_value,
                                 ISEQ_BODY(parent)->location.label,
                                 fname, Qnil, line,
                                 parent, isolated_depth);
@@ -1707,7 +1832,7 @@ eval_string_with_cref(VALUE self, VALUE src, rb_cref_t *cref, VALUE file, int li
     block.as.captured.code.iseq = cfp->iseq;
     block.type = block_type_iseq;
 
-    iseq = eval_make_iseq(src, file, line, NULL, &block);
+    iseq = eval_make_iseq(src, file, line, &block);
     if (!iseq) {
         rb_exc_raise(ec->errinfo);
     }
@@ -1728,7 +1853,7 @@ eval_string_with_scope(VALUE scope, VALUE src, VALUE file, int line)
 {
     rb_execution_context_t *ec = GET_EC();
     rb_binding_t *bind = Check_TypedStruct(scope, &ruby_binding_data_type);
-    const rb_iseq_t *iseq = eval_make_iseq(src, file, line, bind, &bind->block);
+    const rb_iseq_t *iseq = eval_make_iseq(src, file, line, &bind->block);
     if (!iseq) {
         rb_exc_raise(ec->errinfo);
     }
@@ -1770,7 +1895,7 @@ rb_f_eval(int argc, const VALUE *argv, VALUE self)
     int line = 1;
 
     rb_scan_args(argc, argv, "13", &src, &scope, &vfile, &vline);
-    SafeStringValue(src);
+    StringValue(src);
     if (argc >= 3) {
         StringValue(vfile);
     }
@@ -1948,13 +2073,14 @@ rb_yield_refine_block(VALUE refinement, VALUE refinements)
     else {
         const struct rb_captured_block *captured = VM_BH_TO_ISEQ_BLOCK(block_handler);
         struct rb_captured_block new_captured = *captured;
+        const VALUE *const argv = &new_captured.self; /* dummy to suppress nonnull warning from gcc */
         VALUE new_block_handler = VM_BH_FROM_ISEQ_BLOCK(&new_captured);
         const VALUE *ep = captured->ep;
         rb_cref_t *cref = vm_cref_push(ec, refinement, ep, TRUE, FALSE);
         CREF_REFINEMENTS_SET(cref, refinements);
         VM_FORCE_WRITE_SPECIAL_CONST(&VM_CF_LEP(ec->cfp)[VM_ENV_DATA_INDEX_SPECVAL], new_block_handler);
         new_captured.self = refinement;
-        return vm_yield_with_cref(ec, 0, NULL, RB_NO_KEYWORDS, cref, FALSE);
+        return vm_yield_with_cref(ec, 0, argv, RB_NO_KEYWORDS, cref, FALSE);
     }
 }
 
@@ -1963,7 +2089,7 @@ static VALUE
 eval_under(VALUE self, int singleton, VALUE src, VALUE file, int line)
 {
     rb_cref_t *cref = vm_cref_push(GET_EC(), self, NULL, FALSE, singleton);
-    SafeStringValue(src);
+    StringValue(src);
 
     return eval_string_with_cref(self, src, cref, file, line);
 }
@@ -1976,19 +2102,24 @@ specific_eval(int argc, const VALUE *argv, VALUE self, int singleton, int kw_spl
         return yield_under(self, singleton, 1, &self, kw_splat);
     }
     else {
-        VALUE file = Qundef;
+        VALUE file = Qnil;
         int line = 1;
         VALUE code;
 
         rb_check_arity(argc, 1, 3);
         code = argv[0];
-        SafeStringValue(code);
+        StringValue(code);
         if (argc > 2)
             line = NUM2INT(argv[2]);
         if (argc > 1) {
             file = argv[1];
             if (!NIL_P(file)) StringValue(file);
         }
+
+        if (NIL_P(file)) {
+            file = get_eval_default_path();
+        }
+
         return eval_under(self, singleton, code, file, line);
     }
 }
@@ -2515,9 +2646,18 @@ rb_current_realfilepath(void)
         if (path == eval_default_path) {
             return Qnil;
         }
-        else {
-            return path;
+
+        // [Feature #19755] implicit eval location is "(eval at #{__FILE__}:#{__LINE__})"
+        const long len = RSTRING_LEN(path);
+        if (len > EVAL_LOCATION_MARK_LEN+1) {
+            const char *const ptr = RSTRING_PTR(path);
+            if (ptr[len - 1] == ')' &&
+                memcmp(ptr, "("EVAL_LOCATION_MARK, EVAL_LOCATION_MARK_LEN+1) == 0) {
+                return Qnil;
+            }
         }
+
+        return path;
     }
     return Qnil;
 }
